print("import: 'passlib'")
import passlib

print("import: 'passlib._setup'")
import passlib._setup

print("import: 'passlib.crypto'")
import passlib.crypto

print("import: 'passlib.crypto._blowfish'")
import passlib.crypto._blowfish

print("import: 'passlib.crypto.scrypt'")
import passlib.crypto.scrypt

print("import: 'passlib.ext'")
import passlib.ext

print("import: 'passlib.ext.django'")
import passlib.ext.django

print("import: 'passlib.handlers'")
import passlib.handlers

print("import: 'passlib.tests'")
import passlib.tests

print("import: 'passlib.utils'")
import passlib.utils

print("import: 'passlib.utils.compat'")
import passlib.utils.compat

